#!/usr/bin/perl


use strict;
use File::Path qw(mkpath);
use File::Copy;
use File::Basename;
use Encode;
use Cwd;

my $program = basename($0);

use Getopt::Long;
my ($dir, $encod, $file, $lang, $process, $stop_list);
eval {
      $SIG{__WARN__} = sub {usage(1)};
      GetOptions("directory=s" => \$dir,
                 "encoding=s"  => \$encod,
		 "file=s"      => \$file,
		 "language=s"  => \$lang,
		 "process=s"   => \$process, 
		 "stop_list=s" => \$stop_list, # Not used for the time being
		);
     };
# Resetting the WARN handler
$SIG{__WARN__} = sub {warn($_[0])};

if ( ! $file or ! $lang ) {
	usage(1);
	}

if ( ! -f $file ) {
	print STDERR "Error: file \"$file\" does not exist\n";
	exit (2);
	}

if ( ! $process ) {
	$process = "a";
	}
elsif ( $process !~ /^(a|t|e)\z/i ) {
	print STDERR "Error: wrong option \"$process\"\n";
	usage(2);
	}

if ( $encod ) {
	if ( $encod =~ /^utf\W?8\z/i ) {
		$encod = "utf8";
		}
	elsif ( $encod =~ /^iso\W?(lat(in)?|8859)\W?1\z/i ) {
		$encod = "iso-8859-1";
		}
	else	{
		print STDERR "Error: wrong encoding \"$encod\"\n";
		usage(1);
		}
	}
else	{
	$encod = "utf8";
	}
my $enc = find_encoding($encod);	#  Encoding object (see Encode module)

# Variables used by function "category"
my ($CatNom, $CatAdj, $CatNP);

# Variables used by function "tokenize"
my ($apos, $LClitic, $RClitic);
my $LChar = qr/[[{(\\`"‚„†‡‹‘’“”•–—›]/o;
my $RChar = qr/[]}\'\`\"),;:\!\?\%‚„…†‡‰‹‘’“”•–—›]/o;
my %exception;

if ( $encod eq "utf8" ) {
	$apos = "(\'|\x{B4}|\x{2019}|\x{2032})";
	}
else	{
	$apos = "(\'|\x{92}|\x{B4})";
	}

if ( $lang =~ /^en\z/i ) {
	$lang = "en";
	$CatNom = "NN";
	$CatAdj = "JJ";
	$CatNP = "NP";
	$RClitic = qr/\'(s|re|ve|d|m|em|ll)|n\'t/;
	%exception = load_exception("ENGLISH");
	}
elsif ( $lang =~ /^fr\z/i ) {
	$lang = "fr";
	$CatNom="NOM";
	$CatAdj="ADJ";
	$CatNP="NPR";
	$LClitic = qr/(?:[CcDdJjLlMmNnSsTt]|[Ee]ntr|[Jj]usqu|[Ll]orsqu|[Qq]u(elqu)?)\'/o;
	$RClitic = qr/-(?:(?:t-)?(?:elles?|ils?|on)|ce|ci|en|je|la|les?|mêmes?|moi|nous|on|toi|tu|vous|y)|[mt]\'/o;
	%exception = load_exception("FRANCAIS");
	}
else	{
	print STDERR "Error: wrong language \"$lang\"\n";
	usage(2);
	}

my $fastrpath = $ENV{'FASTRPATH'};
if ( ! $fastrpath ) {
	print STDERR "Error: environment variable FASTRPATH not defined\n";
	usage(2);
	}
my $fastrlib = "$fastrpath/lib";
my $fastrbin = "$fastrpath/bin";
my $fastrbases = "$fastrpath/bases";
my $fastrtmp = ".";
if ( $ENV{'TMPDIR'} ) {
	$fastrtmp = $ENV{'TMPDIR'};
	}
$ENV{'FASTRLIB'} = $fastrlib;
$ENV{'FASTRBIN'} = $fastrbin;
$ENV{'FASTRBASES'} = $fastrbases;
$ENV{'FASTRTMP'} = $fastrtmp;

if ( ! $dir ) {
	$dir = "$fastrbases";
	}
# Normalisation to the absolute path
$dir = Cwd::abs_path($dir);

my ($base, $path) = fileparse($file);
if ( $base =~ /^([-.\w]+?)(\.txt)?(\.(g?z|Z|bz2))?\z/ ) {
	$base = $1;
	}
else	{
	print STDERR "Error: wrong name for file \"$base\"\n";
	exit (2);
	}

my $treetagger = "$fastrbin/tree-tagger -token -lemma -sgml";

if ( $process !~ /e/i ) {
	# Preparation of the base repository
	prepare_base($base, $dir);
	copy("$fastrlib/Fastr.lang-$lang", "$dir/$base/config/Fastr.lang-$lang") or die "$!,";
	# Copy and modification of the config file
	open(CFG, "$fastrlib/Fastr.conf-$lang") or die "$!,";
	open(OUT, "> $dir/$base/config/Fastr.conf-$lang") or die "$!,";
	while(<CFG>) {
		if ( /^Dictionary file/ ) {
			s|\$\{FASTRTMP\}/.*|$dir/$base/bin/$base\'|;
			}
		elsif ( /^Language file/ ) {
			s|\$\{FASTRBASES\}/.*|$dir/$base/config/Fastr.lang-$lang\'|;
			}
		elsif ( /^Log file/ ) {
			s|\$\{FASTRTMP\}/||;
			}
		print OUT;
		}
	close CFG;
	close OUT;

	if ( $file =~ /\.(g?z|Z)\z/ ) {
		open(INP, "gzip -c -d $file |") or die "$!,";
		}
	elsif ( $file =~ /\.bz2\z/ ) {
		open(INP, "bzip2 -c -d $file |") or die "$!,";
		}
	else	{
		open(INP, $file) or die "$!,";
		}
	
	# Reading the terms file
	my (%list, %syn);
	while(<INP>) {
		next if /^\s*$/o or /^#/o;
		chomp;
		my $str = $enc->decode($_, Encode::FB_QUIET);
		if ( $_ ) {
			print STDERR "Error: wrong character for encoding \"$encod\" in file \"$file\"\n";
			exit (2);
			}
		if ( $str =~ /\s+=>\s+/ ) {
			$list{$'} ++;
			push(@{$syn{$'}}, $`);
			}
		else	{
			$list{$str} ++;
			}
		}
	close INP;
	# Making a list of it and saving it
	my @list;
	open(OUT, "| gzip -c -9 > $dir/$base/tables/$base.TAB.z") or die "$!,";
	my $num = 0;
	my $tag1 = "p" . substr($lang, 0, 1);
	my $tag2 = "sy" . substr($lang, 0, 1);
	foreach my $term (sort keys %list) {
		push(@list, $term);
		my $cterm = $enc->encode($term);
		printf OUT "%06d\t<$tag1>%s</$tag1>\n", $num, $cterm;
		$num ++;
		if ( $syn{$term} ) {
			foreach my $syn (sort @{$syn{$term}}) {
				push(@list, $syn);
				my $csyn = $enc->encode($syn);
				printf OUT "%06d\t<$tag1>%s</$tag1>", $num, $cterm;
				print  OUT "<$tag2>$csyn</$tag2>\n";
				$num ++;
				}
			}
		}
	
	open(OUT, "> $fastrtmp/terms$$") or die "$!,";
	#
	# Processing English terms ...
	#
	if ( $lang eq "en" ) {
		foreach my $term (@list) {
			my $simple = tokenize($term);
			# We add a final period unless there is an exclamation or interrogation mark
			$simple =~ s/(?<![\!\?])[.\s]*\z/ ./o;
			print OUT $enc->encode($simple) . "\n";
			}
		close OUT;
		# Let's clean memory
		@list = ();
		my @lines;
		if ( $encod eq "iso-8859-1" ) {
			open(INP, "$treetagger $fastrlib/english.par $fastrtmp/terms$$ |") or die "$!,";
			}
		elsif ( $encod eq "utf8" ) {
			open(INP, "$treetagger $fastrlib/english.par $fastrtmp/terms$$ |") or die "$!,";
			}
		while(<INP>) {
			chomp;
			push(@lines, $enc->decode($_));
			}
		close INP;
		category(\@lines, $lang);
		open(OUT, ">  $fastrtmp/$base.tt") or die "$!,";
		open(TAG, "> $dir/$base/rules/$base.tag") or die "$!,";
		foreach my $line (@lines) {
			print OUT $enc->encode("$line\n");
			$line =~ s/@@@/\./go;
			$line =~ s/^([a-z]\.)\t.+\t/$1\tJJ\t/go;
			print TAG $enc->encode("$line\n"); # End of line?
			}
		close OUT;
		close TAG;
		}
	#
	# or French terms.
	#
	elsif ( $lang eq "fr" ) {
	#	binmode(STDOUT, ":utf8");
		foreach my $term (@list) {
			my $simple = tokenize($term);
			$simple =~ s/[.\s]+\z//o;
			$simple =~ s/\./\@\@\@/go;
			$simple =~ s|/| §§SLASH§§ |go;
			$simple =~ s/ ([A-Z]+)\z/ $1§§FIN§§/o;
			$simple =~ s/: /:/go;	# ???
			# We add a final period unless there is an exclamation or interrogation mark
			$simple =~ s/(?<![\!\?])\z/\n./o;
			print OUT $enc->encode($simple) . "\n";
			}
		close OUT;
		my @lines;
		if ( $encod eq "iso-8859-1" ) {
			open(INP,  "$treetagger $fastrlib/french.par $fastrtmp/terms$$ |") or die "$!,";
			}
		elsif ( $encod eq "utf8" ) {
			open(INP,  "$treetagger $fastrlib/french-utf8.par $fastrtmp/terms$$ |") or die "$!,";
			}
		while(<INP>) {
			chomp;
			push(@lines, $enc->decode($_));
			}
		close INP;
		category(\@lines, $lang);
		open(OUT, ">  $fastrtmp/$base.tt") or die "$!,";
		foreach my $line (@lines) {
			$line =~ s/@@@/\./g;
			$line =~ s/§§FIN§§//g;
			$line =~ s/^([a-z]\.)\t.+\t/$1\tADJ:_:_\t/g;
			print OUT $enc->encode("$line\n"); # End of line?
			}
		close OUT;
		if ( $encod eq "iso-8859-1" ) {
			system "flemm -entree $fastrtmp/$base.tt -sortie $fastrtmp/$base.flem -tagger treetagger --isolat1";
			}
		elsif ( $encod eq "utf8" ) {
			system "flemm -entree $fastrtmp/$base.tt -sortie $fastrtmp/$base.flem -tagger treetagger --utf8";
			}
		my %tag2tag = init_table();
		open(INP, "$fastrtmp/$base.flem") or die "$!,";
		open(TAG, "> $dir/$base/rules/$base.tag") or die "$!,";
		while(<INP>) {
			next if /^\s*$/;
			s/^\s+//o;
			s/\s+$//o;
			s/ +/ /go;
			s/ *\t */\t/go;
			s|^§§SLASH§§\t.+?\t§§slash§§|/\tSYM\t/|gio;
			s/^([^\t]+\t[^\t]+\t[^\|]+?)\s*\|.*/$1/go;
			my $line = reductagsflemm($enc->decode($_));
			$line = tag2tag($line, \%tag2tag);
			print TAG $enc->encode($line), "\n";
			}
		close INP;
		close TAG;
		}
	} # process != "e"
	
if ( $process !~ /t/i ) {
	my @lines;
	open(INP, "$dir/$base/rules/$base.tag") or die "$!,";
	while(<INP>) {
		chomp;
		push(@lines, $enc->decode($_));
		}
	close INP;

	open(TRM, "> $dir/$base/rules/$base.trm") or die "$!,";
	my $nb = 0;
	my (@lemmas, @tags, @terms);
	foreach my $line (@lines) {
		# This code comes from "TreeTaggertoTerm" 
		# Copyright (C) 1999 Christian Jacquemin, LIMSI-CNRS
		#  with some adapations.
		
		# splitting the line into: inflected word, tag, and lemma
		my ($INFLWORD, $TAG, $LEMMA) = split(/\t/, $line);
    
		# unknown lemmas are replaced by the inflected word
		if ( $LEMMA eq "<unknown>" or $LEMMA eq "\@card\@" ) {
			$LEMMA = $INFLWORD;
			}
    
		# normalizing words
		if ( $TAG !~ /[\#\$\'()\.:;,]/ ) {
			# suppression of parasite characters
			$LEMMA =~ s/[\"\{\}\$\~\=\|*,\\\]\[\`\<\>\?\@\_]//g;
			# deletion of double quotes
			$LEMMA =~ s/\'\'//g;
			# deletion of initial and final single quote
			$LEMMA =~ s/^\'//;
			$LEMMA =~ s/\'$//;
			# normalization of double or triple hyphens
			$LEMMA =~ s/---/-/g;
			$LEMMA =~ s/--/-/g;
			# deletion of initial and final  hyphen
			$LEMMA =~ s/^-//;
			$LEMMA =~ s/-$//;
			}

		# dot finishing a term: prints the term
		if( $TAG eq "SENT" or $TAG eq "PON:sep" ) {
			my $term = join(" ", @lemmas) . "\t" . join(" ", @tags);
			$term = sprintf "%06d\t%s", $nb++, $term;
			print TRM $enc->encode($term), "\n";
			push(@terms, $term);
			@lemmas = ();
			@tags = ();
			}
		else	{
			push( @tags , $TAG );
			push( @lemmas , $LEMMA );
			}
		}
	close TRM;
	
	#
	# Processing English terms ...
	#
	if ( $lang eq "en" ) {
		my $der_families = "der-families-en";
		$der_families .= ".utf8" if $encod eq "utf8";
		my $tmpfile = lex2basemorpho("", $der_families);
		my @rules = terms2rules(\@terms, "$fastrlib/TAGS-TreeTagger-en");
		
		# Rules on words
		my @words;
		my $oldword;
		foreach my $word (sort grep(/^Word /, @rules)) {
			next if $word eq $oldword;
			$oldword = $word;
			push @words, $word;
			}
		word2families($tmpfile, \@words, "$fastrlib/sem-classes-en");
		$oldword = "";
		open(RW, "> $dir/$base/rules/$base.R.w") or die "$!,";
		foreach my $word (sort grep /^Word/, @words) {
			next if $word eq $oldword;
			$oldword = $word;
			print RW $enc->encode($word), "\n";
			}
		close RW;
		
		# Rest of the rules
		open(RT, "> $dir/$base/rules/$base.R.t") or die "$!,";
		foreach my $rule (@rules) {
			next if $rule =~ /^Word /o or $rule =~ /^\s*\z/o;
			print RT $enc->encode($rule), "\n";
			}
		close RT;
		
		# Initialisation of Fastr
		system "$fastrbin/Fastr -C $dir/$base/config/Fastr.conf-en -z";
		
		# Compilation of rules on words
		system "$fastrbin/Fastr -C $dir/$base/config/Fastr.conf-en -c $dir/$base/rules/$base.R.w";
		
		# Compilation of rules on terms
		system "$fastrbin/Fastr -C $dir/$base/config/Fastr.conf-en -c $dir/$base/rules/$base.R.t";
		}

	#
	# or French terms.
	#
	elsif ( $lang eq "fr" ) {
		sortlexicon(\@lines, "$fastrtmp/$base.tri");
		my $der_families = "der-families-fr";
		$der_families .= ".utf8" if $encod eq "utf8";
		my $tmpfile = lex2basemorpho("$fastrtmp/$base.tri", $der_families);
		my @rules = terms2rules(\@terms, "$fastrlib/TAGS-TreeTagger-fr");
		
		# Rules on words
		my @words;
		my $oldword;
		foreach my $word (sort grep(/^Word /, @rules)) {
			next if $word eq $oldword;
			$oldword = $word;
			push @words, $word;
			}
		word2families($tmpfile, \@words, "$fastrlib/sem-classes-fr");
		$oldword = "";
		open(RW, "> $dir/$base/rules/$base.R.w") or die "$!,";
		foreach my $word (sort grep /^Word/, @words) {
			next if $word eq $oldword;
			$oldword = $word;
			print RW $enc->encode($word), "\n";
			}
		close RW;
		
		# Rest of the rules
		open(RT, "> $dir/$base/rules/$base.R.t") or die "$!,";
		foreach my $rule (@rules) {
			next if $rule =~/^Word /o or $rule =~ /^\s*\z/o;
			print RT $enc->encode($rule), "\n";
			}
		close RT;
		
		# Initialisation of Fastr
		system "$fastrbin/Fastr -C $dir/$base/config/Fastr.conf-fr -z";
		
		# Compilation of rules on words
		system "$fastrbin/Fastr -C $dir/$base/config/Fastr.conf-fr -c $dir/$base/rules/$base.R.w";
		move("Fastr.log", "Fastr.log.w");

		# Compilation of rules on terms
		system "$fastrbin/Fastr -C $dir/$base/config/Fastr.conf-fr -c $dir/$base/rules/$base.R.t";
		move("Fastr.log", "Fastr.log.t");
		}
	}

exit 0;


sub usage
{
my $cr = shift;

print STDERR <<EOT;

Usage: $program -f file -l (en|fr) [ -p (a|t|e) ] [ -d directory ] [ -e (utf8|isolat1) ]

Options: 
        -f, --file=FILE                 Input file, 
        -l, --language=(en|fr)          Language: English (en) or French (fr), 
	-d, --directory=DIR             Install the base in directory DIR, 
	-e, --encoding=(utf8|isolat1)   Encoding of input file: UTF-8 or ISO-Latin1 
	                                (UTF-8 is the default), 
        -p, --process=(a|t|e)           Process: all (a), only up to the tags (t) 
                                        or resume from the tagfile to the end (e), 
                                        (all is the default). 

EOT

exit ($cr);				
}

sub load_exception
{
my $language = shift;
my %hash;
my $ok = 0;

while(<DATA>) {
	next if /^\s*\z/o;
	next if /^#/o;
	if ( /^%% $language/o ) {
		$ok = 1;
		}
	elsif ( /^%% END/o ) {
		last;
		}
	elsif ( /^%% / ) {
		$ok = 0;
		}
	elsif ( $ok ) {
		chomp;
		$hash{$_} ++;
		}
	}

return %hash;
}

sub prepare_base
{
my $base = shift;
my $dir = shift;

if ( ! -d "$dir/$base" ) {
	mkpath("$dir/$base") or die "Error: cannot make path \"$dir/$base\"";
	}
for my $subdir (qw(bin config log rules tables)) {
	my $current = "$dir/$base/$subdir";
	if ( -d $current ) {
		opendir(DIR, $current) or die "$!,";
		my @files = grep {not /^\.{1,2}\z/} readdir DIR;
		closedir DIR;
		foreach my $file (@files) {
			unlink "$current/$file" or die "$subdir/$file: $!,";
			}
		}
	else	{
		mkpath($current) or die "$current: $!,";
		}
	}
}


sub reduct
{
my $term = shift;

$term =~ s/\+/ sgnp /go;
$term =~ s/, / /go;
$term =~ s|[][(){}]+| |go;
# Apostrophes (quotes) are kept only after a letter
$term =~ s/(?<=\p{IsAlpha})'/' /go;
$term =~ s/(?<=\P{IsAlpha})'/ /go;
$term =~ s/[\p{IsZs}]+/ /go;
$term =~ s/([^\p{IsAlnum}.:;'!\? ])/substitute($1)/ge;
$term =~ s/^ +//o;
$term =~ s/ +\z//o;
$term =~ s/ +/ /og;

return lc($term);
}

sub tokenize
{
my $term = shift;

$term =~ s/\+/ sgnp /go;
# Replacing every kind of apostrophe by the simple quote
$term =~ s/$apos/\'/go;
$term =~ s/(\.\.\.)/ ... /g;
$term =~ s/([;\!\?])([^ ])/$1 $2/g;
$term =~ s/([.,:])([^ 0-9.])/$1 $2/g;
# Why do we suppress these separators?
$term =~ s|[][(){}]+| |go;
$term =~ s/[\p{IsZs}]+/ /go;
$term =~ s/([^\p{IsAlnum}.:;'!\? ])/substitute($1)/ge;
$term =~ s/^ +//o;
$term =~ s/ +\z//o;
$term =~ s/ +/ /og;

my @words = split(/\s+/, $term);
foreach my $word (@words) {
	$word =~ s/^($LChar)(.)/$1\n$2/o;
	$word =~ s/(.)($RChar)\z/$1\n$2/o;
	my @items = ();
	foreach my $item (split(/\n/, $word)) {
		my $clitic;
		if ( ! $exception{$item} ) {
			if ( $LClitic and $item =~ /^($LClitic)/o ) {
				push @items, $1;
				$item = $';
				}
			if ( $RClitic and $item =~ /($RClitic)\z/o ) {
				$clitic = $1;
				$item = $`;
				}
			$item =~ s/\'/\n/go;
			}
		push @items, $item;
		push @items, $clitic if $clitic;
		}
	$word = join("\n", @items);
	}
$term = join("\n", @words);

return lc($term);
}

sub substitute
{
my $char = shift;

if ( $char =~ /^[\p{IsS}\p{IsPd}]\z/o ) {
	return " ";
	}
elsif ( $char =~ /^\p{IsP}\z/o ) {
	return "";
	}
}

sub category
{
my ($lines, $lang) = @_;

if ( $lang eq "en" ) {
	for ( my $nb = 1 ; $nb < $#{$lines} ; $nb ++ ) {
		if ( $lines->[$nb] =~ /\tVBG\t/ and
		     $lines->[$nb - 1] =~ /^\.\tSENT\t\.$/ and
		     $lines->[$nb + 1] =~ /^\.\tSENT\t\.$/ ) {
			$lines->[$nb] =~ s/^(\S+)\tVBG\t\S+$/$1\tNN\t$1/;
			}
		}
	}
elsif ( $lang eq "fr" ) {
	for ( my $nb = 1 ; $nb < $#{$lines} ; $nb ++ ) {
		if ( $lines->[$nb] =~ /\tVER:ppre\t/ and
		     $lines->[$nb - 1] =~ /^\.\tPON:sep\t\.$/ and
		     $lines->[$nb + 1] =~ /^\.\tPON:sep\t\.$/ ) {
			$lines->[$nb] =~ s/^(\S+)\tVER:ppre\t\S+$/$1\tNOM\t$1/;
			}
		}
	}

foreach (@{$lines}) {
	chomp;
	my ($Terme, $Cat, $Lem) = split(/\t/);
#	if ( $tab{$Terme} ) {
#		$Cat = $CatNP;
#		$Lem = $Terme;
#		}
#	elsif ( $Cat eq "VBD" ) {
	if ( $Cat eq "VBD" ) {
		$Cat = $CatAdj;
		$Lem = $Terme;
		}
	elsif ( $Cat eq "VER:pres" ) {
		$Cat = $CatNom;
		$Lem = $Terme;
		}
	elsif ( $Cat eq "VER:simp" ) {
		$Cat = $CatNom;
		$Lem = $Terme;
		}
	elsif ( $Cat eq "VER:aux:pres" ) {
		$Cat = $CatNom;
		$Lem = $Terme;
		}						
	elsif ( $Lem eq "<unknown>" ) {
		$Lem = $Terme;
		if ( $lang eq "en" ) {
			if ( $Terme  =~ /^.*ic\z/o ) {
				$Cat = $CatAdj;
				}
			elsif ( $Terme  =~ /^.*ics\z/o ) {
				$Cat = $CatNom;
				}
			elsif ( $Terme =~ /^.*[ai]ble\z/o ) {
				$Cat = $CatAdj;
				}
			elsif ( $Terme =~ /^.*ian\z/o ) {
				$Cat = $CatAdj;
				}
			elsif ( $Terme =~ /^.*[ao]ry\z/o ) {
				$Cat = $CatAdj;
				}
			elsif ( $Terme =~ /^.*ive\z/o ) {
				$Cat = $CatAdj;
				}
			elsif ( $Terme =~ /^.*al\z/o ) {
				$Cat = $CatAdj;
			}
		}
		else	{
			$Cat = $CatNom;
			}
		}
	$_ = join("\t", $Terme, $Cat, $Lem);
	}
}

sub reductagsflemm
{
my $line = shift;

my ($flex, $tag, $lem) = split ("\t", $line);
return $line if $tag !~ /(DET|PRE|PRO|VER)/o;

if ($tag =~ /DET/) {
	$tag =~ s/DET\(def\).+/DET:def/o;
	$tag =~ s/DET\(indef\).+/DET:indef/o;
	}

elsif ($tag =~ /PRE/) {
	$tag =~ s/PRE\(det\).+/PRE:det/o;
	}

elsif ($tag =~ /PRO/) {
	$tag =~ s/PRO\(clit\).+/PRO:clit/o;
	$tag =~ s/PRO\(demo:attr\).+/PRO:demo:attr/o;
	$tag =~ s/PRO\(demo:pred\).+/PRO:demo:pred/o;
	$tag =~ s/PRO\(indef:attr\).+/PRO:indef:attr/o;
	$tag =~ s/PRO\(indef:pred\).+/PRO:indef:pred/o;
	$tag =~ s/PRO\(inter\).+/PRO:inter/o;
	$tag =~ s/PRO\(pers:clit\).+/PRO:pers:clit/o;
	$tag =~ s/PRO\(pers:conj\).+/PRO:pers:conj/o;
	$tag =~ s/PRO\(pers:disj\).+/PRO:pers:disj/o;
	$tag =~ s/PRO\(pers:refl\).+/PRO:pers:refl/o;
	$tag =~ s/PRO\(poss\).+/PRO:poss/o;
	$tag =~ s/PRO\(rela\).+/PRO:rela/o;
	}

elsif ($tag =~ /VER/) {
	$tag =~ s/VER\(aux:cond\).+/VER:aux:cond/o;
	$tag =~ s/VER\(aux:futu\).+/VER:aux:futu/o;
	$tag =~ s/VER\(aux:impf\).+/VER:aux:impf/o;
	$tag =~ s/VER\(aux:pper\).+/VER:aux:pper/o;
	$tag =~ s/VER\(aux:ppre\).+/VER:aux:ppre/o;
	$tag =~ s/VER\(aux:pres\).+/VER:aux:pres/o;
	$tag =~ s/VER\(aux:simp\).+/VER:aux:simp/o;
	$tag =~ s/VER\(aux:subi\).+/VER:aux:subi/o;
	$tag =~ s/VER\(aux:subp\).+/VER:aux:subp/o;
	$tag =~ s/VER:aux:cond.+/VER:aux:cond/o;
	$tag =~ s/VER:aux:futu.+/VER:aux:futu/o;
	$tag =~ s/VER:aux:impf.+/VER:aux:impf/o;
	$tag =~ s/VER:aux:pper.+/VER:aux:pper/o;
	$tag =~ s/VER:aux:ppre.+/VER:aux:ppre/o;
	$tag =~ s/VER:aux:pres.+/VER:aux:pres/o;
	$tag =~ s/VER:aux:simp.+/VER:aux:simp/o;
	$tag =~ s/VER:aux:subi.+/VER:aux:subi/o;
	$tag =~ s/VER:aux:subp.+/VER:aux:subp/o;
	$tag =~ s/VER\(cond\).+/VER:cond/o;
	$tag =~ s/VER\(futu\).+/VER:futu/o;
	$tag =~ s/VER\(impf\).+/VER:impf/o;
	$tag =~ s/VER\(pper\).+/VER:pper/o;
	$tag =~ s/VER\(ppre\).+/VER:ppre/o;
	$tag =~ s/VER\(pres\).+/VER:pres/o;
	$tag =~ s/VER\(simp\).+/VER:simp/o;
	$tag =~ s/VER\(subi\).+/VER:subi/o;
	$tag =~ s/VER\(subp\).+/VER:subp/o;
	$tag =~ s/VER:cond.+/VER:cond/o;
	$tag =~ s/VER:futu.+/VER:futu/o;
	$tag =~ s/VER:impf.+/VER:impf/o;
	$tag =~ s/VER:pper.+/VER:pper/o;
	$tag =~ s/VER:ppre.+/VER:ppre/o;
	$tag =~ s/VER:pres.+/VER:pres/o;
	$tag =~ s/VER:simp.+/VER:simp/o;
	$tag =~ s/VER:subi.+/VER:subi/o;
	$tag =~ s/VER:subp.+/VER:subp/o;

	$lem =~ s/^([^:]+):.+/$1/;
	}

return "$flex\t$tag\t$lem";
}

sub init_table
{
my %table;

open(TAG, "$fastrlib/new_tag2tag.txt") or die "$!,";
while(<TAG>) {
	chomp;
	my @tags = split(/\t/);
	my $tt_tag = shift @tags;
	foreach my $tag (@tags) {
		last if $tag =~ /^\s*#/o;
		next if $tag =~ /^\s*\z/o;
		my @new_tags = ($tag);
		while( $new_tags[0] =~ /\[(.+?)\]/o ) {
			my @chars = split(//, $1);
			my @tmp_tags;
			foreach my $new_tag (@new_tags) {
				foreach my $char (@chars) {
					my $tmp_tag = $new_tag;
					$tmp_tag =~ s/\[.+?\]/$char/;
					push(@tmp_tags, $tmp_tag);
					}
				}
			@new_tags = @tmp_tags;
			}
		foreach my $new_tag (@new_tags) {
			$table{$new_tag} = $tt_tag;
			if ( $new_tag =~ /^(.+?)\(.+?\)(:.*)/ ) {
				$table{"$1$2"} = $tt_tag;
				}
			}
		}
	}
close TAG;

return %table;
}

sub tag2tag
{
my ($line, $tabref) = @_;

my ($term, $tag, $lemma) = split(/\t/, $line);
if ( exists $tabref->{$tag} ) {
	$line = "$term\t$tabref->{$tag}\t$lemma";
	}
else	{
	print STDERR "Warning: tag \"$tag\" not in the conversion table\n";
	}

return $line;
}

sub sortlexicon
{
my ($rlines, $file) = @_;

my ($term, $synpat);
my (@terms, @synpats);
my (%listCat, %listIdt);

# If we decide to use an antilexicon, it should be done here!

my $nb = 1;
foreach my $line (@{$rlines}) {
	# selecting first answer from flemm if several are given
	$line =~ s/^(.+?\t;+?\t.+?)\s*||/$1/o;
	my ($flex, $cat, $lem) = split(/\t/, $line);
	$lem =~ s/^([^:]+):.+/$1/og;
	
	if ($cat eq "PON:sep") {
		push(@terms, $term);
		push(@synpats, $synpat);
		$term = $synpat = "";
		$nb ++;
		}
	else	{
		# Normalizing the tags
		$cat = "N"    if $cat =~ /NOM/o;
		$cat = "A"    if $cat =~ /ADJ/o;
		$cat = "V"    if $cat =~ /VER/o;
		$cat = "P"    if $cat =~ /PRE/o;
		$cat = "D"    if $cat =~ /DET/o;
		$cat = "Np"   if $cat =~ /NPR/o;
		$cat = "Adv"  if $cat =~ /ADV/o;
		$cat = "Abrv" if $cat =~ /ABR/o;
		$cat = "Conj" if $cat =~ /CON/o;
		$cat = "Ponc" if $cat =~ /PON/o;
		$cat = "Pro"  if $cat =~ /PRO/o;
		$cat = "Int"  if $cat =~ /INT/o;
		$cat = "Sym"  if $cat =~ /SYM/o;
		$term .= " " if $term;
		$synpat .= " " if $synpat;
		$term .= $flex;
		$synpat .= $cat;
		
		if ( $listCat{$flex} ) {
			$listCat{$flex} .= " $cat";
			$listIdt{$flex} .= " $nb";
			}
		else	{
			if ( $flex !~ /[(){}$%&+=<>*",;:!?0-9]/o ) {
				$listCat{$flex} = $cat;
				$listIdt{$flex} = $nb;
				}
			}
		}
	}

open (TRI, "> $file") or die "$!,";

foreach my $key (keys %listCat) {
	# Skipping meaningless terms
	next if $key =~ /^\W\z/o or $key =~ /^\d+\z/o or $key =~ /^\w\z/o;
	
	my @cats = split(/\s+/, $listCat{$key});
	my @idts = split(/\s+/, $listIdt{$key});
	
	my %tmp;
	foreach my $cat (@cats) {
		$tmp{$cat} ++;
		}
	
	print TRI "\n", $enc->encode($key);
	foreach my $cat (sort keys %tmp) {
		print TRI "\t$cat:$tmp{$cat}";
		}
	print TRI "\n";
	
	foreach my $cat (@cats) {
		my $idt = shift @idts;
		print TRI "\t$cat\t", $enc->encode($terms[$idt - 1]), "\t$synpats[$idt - 1]\n";
		}
	}

close TRI;
}

sub lex2basemorpho
{
my ($lex, $base) = @_;
my $file = "$fastrlib/$base";

my $user = "user";
if ( $ENV{'USER'} ) {
	$user = $ENV{'USER'};
	}
elsif ( $ENV{'USERNAME'} ) {
	$user = $ENV{'USERNAME'};
	}

if ( $user and -f "$fastrtmp/$base.$user" ) {
	my $date_m = (stat("$fastrtmp/$base.$user"))[9];
	if ( $date_m > (stat("$fastrlib/$base"))[9] ) {
		$file = "$fastrtmp/$base.$user";
		}
	}

my %morpho;
open(BMD, $file) or die "$!,";
while(<BMD>) {
	chomp;
	my $line = $enc->decode($_);
	my ($head) = $line =~ /^(.+?):/o;
	$head =~ s/\\.*//o;
	if ( ! exists $morpho{$head} ) {
		$morpho{$head} = $line;
		}
	}
close BMD;

my @words;
if ( $lex ) {
	open(LEX, $lex) or die "$!,";
	while(<LEX>) {
		next unless /^[^\t].+[NA]:/o;
		my $line = $enc->decode($_);
		my ($word) = $line =~ /^\s*([^\t]+)/o;
		push(@words, $word);
		}
	close LEX;
	}

my ($inf, $ppres, $actant, $actant2, $ppas, $ppas2, $adj);
map {
	if ( ! exists $morpho{$_} ) {
		$inf = $ppres = $actant = $actant2 = $ppas = $ppas2 = $adj = $_;
	        if ( $_ =~ /ification\z/o ) {
			$inf =~ s/cation\z/er/o;
			$ppres =~ s/cation\z/ant/o;
			$ppas =~ s/cation\z/é/o;
			$actant =~ s/cation\z/eur/o;
			$actant2 =~ s/cation\z/cateur/o;
			$morpho{$_} = "$inf\\V\\: $_\\N\\-er+cation $inf\\V\\-+ $actant\\N\\-er+eur $actant2\\N\\-er+cateur  $ppres\\V\\-er+ant $ppres\\N\\-er+ant $ppres\\A\\-er+ant $ppas\\V\\-er+é $ppas\\N\\-er+é $ppas\\A\\-er+é";
			}
		elsif ( $_ =~ /ation\z/o ) {
			$inf =~ s/ation\z/er/o;
			$ppres =~ s/ation\z/ant/o;
			$actant =~ s/ation\z/eur/o;
			$actant2 =~ s/ation\z/ateur/o;
			$ppas =~ s/ation\z/é/o;
			$morpho{$_} = "$inf\\V\\: $_\\N\\-er+ation $inf\\V\\-+ $actant\\N\\-er+eur $actant2\\N\\-er+ateur $ppres\\V\\-er+ant $ppres\\N\\-er+ant $ppres\\A\\-er+ant $ppas\\V\\-er+é $ppas\\N\\-er+é $ppas\\A\\-er+é";
			}
		elsif ( $_ =~ /abilité\z/o ) {
			$inf =~ s/abilité\z/er/o;
			$ppres =~ s/abilité\z/ant/o;
			$actant =~ s/abilité\z/eur/o;
			$ppas =~ s/abilité\z/é/o;
			$ppas2 =~ s/abilité\z/abilisé/o;
			$adj =~ s/abilité\z/able/o;
			$morpho{$_} = "$inf\\V\\: $_\\N\\-er+abilité $inf\\V\\-+ $actant\\N\\-er+eur $ppres\\V\\-er+ant $ppres\\N\\-er+ant $ppres\\A\\-er+ant $ppas\\V\\-er+é $ppas\\N\\-er+é $ppas\\A\\-er+é $ppas2\\V\\-er+é $ppas2\\N\\-er+é $ppas2\\A\\-er+é $adj\\N\\-er+able $adj\\A\\-er+able";
			}
		elsif ( $_ =~ /ibilité\z/o ) {
			$inf =~ s/ibilité\z/ibiliser/o;
			$ppas =~ s/ibilité\z/ibilisé/o;
			$adj =~ s/ibilité\z/ible/o;
			$morpho{$_} = "$inf\\V\\: $_\\N\\-er+ibilité $inf\\V\\-+ $ppas\\V\\-er+é $ppas\\N\\-er+é $ppas\\A\\-er+é $adj\\N\\-ibiliser+ible $adj\\A\\-ibiliser+ible";
			}
		elsif ( $_ =~ /é\z/o ) {
			$inf =~ s/é\z/er/o;
			$ppres =~ s/é\z/ant/o;
			$actant =~ s/é\z/eur/o;
			$morpho{$_} = "$inf\\V\\: $_\\N\\-er+é $actant\\N\\-er+eur $inf\\V\\-+ $ppres\\V\\-er+ant $ppres\\N\\-er+ant $ppres\\A\\-er+ant $_\\A\\-er+é";
			}
		elsif ( $_ =~ /ée\z/o ) {
			$inf =~ s/ée\z/er/o;
			$ppres =~ s/ée\z/ant/o;
			$actant =~ s/ée\z/eur/o;
			$morpho{$_} = "$inf\\V\\: $_\\N\\-er+ée $actant\\N\\-er+eur $inf\\V\\-+ $ppres\\V\\-er+ant $ppres\\N\\-er+ant $ppres\\A\\-er+ant $_\\A\\-er+ée";
			}
		elsif ( $_ =~ /age\z/o ) {
			$inf =~ s/age\z/er/o;
			$ppres =~ s/age\z/ant/o;
			$actant =~ s/age\z/eur/o;
			$ppas =~ s/age\z/é/o;
			$morpho{$_} = "$inf\\V\\: $_\\N\\-er+age $actant\\N\\-er+eur $inf\\V\\-+ $ppres\\V\\-er+ant $ppres\\N\\-er+ant $ppres\\A\\-er+ant $ppas\\A\\-er+é $ppas\\V\\-er+é $ppas\\N\\-er+é";
			}
		elsif ( $_ =~ /issant\z/o ) {
			$inf =~ s/issant\z/ir/o;
			$ppres = $_;
			$morpho{$_} = "$inf\\V\\: $inf\\V\\-+ $ppres\\V\\-ir+issant $ppres\\N\\-ir+issant $ppres\\A\\-ir+issant";
			}
		elsif ( $_ =~ /ant\z/o ) {
			$inf =~ s/ant\z/er/o;
			$ppres = $_;
			$actant =~ s/ant\z/eur/o;
			$morpho{$_} = "$inf\\V\\: $_\\N\\-er+ant $actant\\N\\-er+eur $inf\\V\\-+ $ppres\\V\\-er+ant $ppres\\N\\-er+ant $ppres\\A\\-er+ant $_\\A\\-+";
			}
		elsif ( $_ =~ /uit\z/o ) {
			$inf =~ s/it\z/ire/o;
			$ppres =~ s/it\z/isant/o;
			$morpho{$_} = "$inf\\V\\: $_\\N\\-ire+it $_\\A\\-ire+it $_\\V\\-ire+it $inf\\V\\-+ $ppres\\V\\-ire+isant $ppres\\N\\-ire+isant $ppres\\A\\-ire+isant";
			}
		elsif ( $_ =~ /[^ao]issement\z/o ) {
			$inf =~ s/issement\z/ir/o;
			$ppres =~ s/issement\z/issant/o;
			$actant =~ s/issement\z/isseur/o;
			$ppas =~ s/issement\z/i/o;
			$morpho{$_} = "$inf\\V\\: $_\\N\\-ir+issement $inf\\V\\-+ $actant\\N\\-ir+isseur $ppres\\V\\-ir+issant $ppres\\N\\-ir+issant $ppres\\A\\-ir+issant $ppas\\V\\-ir+i $ppas\\N\\-ir+i $ppas\\A\\-ir+i";
			}
		elsif ( $_ =~ /ement\z/o ) {
			$inf =~ s/ement\z/er/o;
			$ppres =~ s/ement\z/ant/o;
			$actant =~ s/ement\z/eur/o;
			$ppas =~ s/ement\z/é/o;
			$morpho{$_} = "$inf\\V\\: $_\\N\\-er+ement $inf\\V\\-+ $actant\\N\\-er+eur $ppres\\V\\-er+ant $ppres\\N\\-er+ant $ppres\\A\\-er+ant $ppas\\V\\-er+é $ppas\\N\\-er+é $ppas\\A\\-er+é";
			}
		else	{
			$morpho{$_} = "$_\\N\\: $_\\N\\-+ $_\\V\\-+ $_\\A\\-+";
			}
		}
	} @words;

if ( $user ) {
	$file = "$base.$user";
	}
else	{
	$file = "$base-$$";
	}
open(BMD, "> $fastrtmp/$file") or die "$!,";
foreach my $morpho (sort keys %morpho) {
	print BMD $enc->encode($morpho{$morpho}), "\n";
	}
close BMD;

if ( -w $fastrlib ) {
	move("$fastrtmp/$file", "$fastrlib/$base") or warn "Cannot move \"$base\",";
	}

$file =~ s/$base/tmp-families/;
open(TMP, "> $fastrtmp/$file") or die "$!,";
foreach my $morpho (sort keys %morpho) {
	$morpho =~ s/^(\p{IsUpper}\p{IsLower})/\l$1/o;
	$morpho =~ s/ (\p{IsUpper}\p{IsLower})/ \l$1/go;
	print TMP $enc->encode($morpho), "\n";
	}
close TMP;

return "$fastrtmp/$file";
}

sub terms2rules
{
# This code comes from "TermtoRules" 
# Copyright (C) 1999 Christian Jacquemin, LIMSI-CNRS
#  with some adapations.

# Array with the resulting rules
my @rules;
# Reference to the list of terms
my $ref = shift; 

# reading the extended tags
my $FILENAMETAGS = shift;
open( FILETags , "<$FILENAMETAGS" ) 
    or  die "Can't open file ($FILENAMETAGS)!" ;

my %TAGEQUIVALENCE = ();
while(<FILETags>) {
	# TreeTagger tag
	chomp;
	my $TAGORIG = $_;
	if( !($_ = <FILETags>) ) {
		last;
		}

	# Fastr tag / features
	chomp;
	my $NEWTAG = $_;

	# comments
	if( !<FILETags> ) {
		last;
		}

	# only keeps the cateogory and discard the features
	# stores the translation
	$NEWTAG =~ s/.*<cat> *= *([A-Za-z]+).*/$1/o;
	$TAGEQUIVALENCE{ $TAGORIG } = $NEWTAG;
	}
close( FILETags );

foreach (@{$ref}) {
#	chomp;
	my ( $LABEL , $LEMMAS , $CATEGORIES ) = split( /\t/ );
	my @TAGGER_CATEGORIES = split( / / , $CATEGORIES );
	my @LEMMAS = split( / / , $LEMMAS );

	my @CATEGORIES = ();
	foreach my $TAG ( @TAGGER_CATEGORIES ) {
		$TAG = $TAGEQUIVALENCE{ $TAG };
		push( @CATEGORIES ,  $TAG );
		}

	# the first substantive == the anchor for lexicalisation
	my $ROWFIRSTNOUN = -1;
	my $ROWFIRSTSUB = -1;
	my $ROWFIRSTX = -1;

	# initialization of the loop 
	my $COUNT = 0;
	# loop for collecting the words and searching the anchor of lexicalization
	foreach my $CAT ( @CATEGORIES ) {
		if ( ( $ROWFIRSTNOUN == -1 ) && ( $CAT eq "N" ) ) {
			$ROWFIRSTNOUN = $COUNT;
			}
		if ( ( $ROWFIRSTSUB == -1 ) &&
		     ( ( $CAT eq "V" ) || ( $CAT eq "Av" ) ||
		       ( $CAT eq "A" ) ) ) {
			$ROWFIRSTSUB = $COUNT;
			}
		if( ( $ROWFIRSTX == -1 ) && ( $CAT eq "X" ) ) {
			$ROWFIRSTX = $COUNT;
			}
		$COUNT++;
		}
	# emergency lexicalisation
	if( ($ROWFIRSTNOUN == -1) && ($ROWFIRSTSUB == -1) ) {
		$ROWFIRSTNOUN = $ROWFIRSTX;   
		if( $ROWFIRSTNOUN == -1 ) {
			$ROWFIRSTNOUN = 0;
			}
		}
	else	{
		if( $ROWFIRSTNOUN == -1 ) {
			$ROWFIRSTNOUN = $ROWFIRSTSUB;
			}
		}

	# printing the lemma rules
	$COUNT = 0;
	foreach my $LEMMA ( @LEMMAS ) {
		my $STEM = $LEMMA;
		$STEM =~ s/\\/\\\\/g;
		$STEM =~ s/\'/\\\'/g;
		push @rules, sprintf 
			"Word '%s' %s: <cat> = %3s."
				, $STEM , " " x ( 30 - length( $STEM ) ) 
					, $CATEGORIES[ $COUNT ];
		$COUNT++;
		}

	# printing the term rule
	my $rule = sprintf "Rule N1 ->";
	$COUNT = 2;
	foreach my $CAT ( @CATEGORIES ) {
		$rule .= sprintf " %s%d" , $CAT , $COUNT;
		$COUNT++;
		}
	$rule .= sprintf ":";
	push @rules, $rule;

	push @rules, sprintf " <N1 lex> = '%s%d'" , 
		$CATEGORIES[ $ROWFIRSTNOUN ] , $ROWFIRSTNOUN + 2;
	push @rules, sprintf " <N1 lab> = '%s'" , $LABEL;
   
	# loop for creating the constraints of the term rule
	$COUNT = 0;
	foreach my $LEMMA ( @LEMMAS ) {
		my $STEM = $LEMMA;
		$STEM =~ s/\\/\\\\/g;
		$STEM =~ s/\'/\\\'/g;
		push @rules, sprintf " <%s%d lem> = '%s'" , 
			$CATEGORIES[ $COUNT ] , $COUNT + 2 , $STEM;
		$COUNT++;
		}
	$rules[$#rules] .= ".";
	}

return @rules;
}

sub word2families
{
# This code comes from "WordtoFamilies" 
# Copyright (C) 1999 Christian Jacquemin, LIMSI-CNRS
#  with some adapations.

my $classes = 1;
my $ref = $_[1];

my %LemmaInWordList;
my %semLink;
my %stem;
# constants
my $Is_morph_file = 0;
my $Is_sem_file = 2;
my $Is_word_in_term = 1;
my $Is_word_in_family = 2;


# remembers the term lemmas
foreach (@{$ref}) {
	if( /Word +\'(.*)\' .*\<cat\> *= *([A-Za-z]+)/ ) {
		my $word = $1;
		my $pos = $2;
		$LemmaInWordList{ $word."#".$pos } = $Is_word_in_term;
		}
	}

# Type $Is_morph_file: morphological families: 
#                      stem: family (stem is in family)
# Type $Is_sem_file: semantic families:
#                      base: family (base is not in family if -links)
#                      base: family (base is in family if -classes)

# type = Is_morph_file (0) or Is_sem_file (2)
foreach my $type ($Is_morph_file, $Is_sem_file) {
	# open the file of families
	my $NAMEFILE = $_[$type];
	open( FILEfamilies , "<$NAMEFILE" ) 
	  or  die "Can't open file $NAMEFILE!" ;

	# selects the families in which one of the lemmas is in a term
	while ( my $line = <FILEfamilies> ) {

		my $lineIni = $line;
		chomp ($line);
		$line = $line . " ";

		# scans the whole family to check whether one of the lemmas is
		# in the lexicon
		my $OneLemmaInLexicon = 0;
		my ($baseMember, $famMembers) = split(":", $line);
		my ($baseMemberword, $baseMemberpos, $famMember) = split(/\\/, $baseMember);

		while ($famMembers =~ m/[^ ]+?\\.+?\\.*? /go) {
			my ($word, $pos, $famMember) =  split /\\/, $&;

			if( $LemmaInWordList{ $word."#".$pos } == $Is_word_in_term ){
			$OneLemmaInLexicon = 1;
				}
			}

		# Morphology: one of the lemmas in this family is found in the lexicon
		# Semantics and "link" mode: the base lemma is found in the lexicon
		# Semantics and "class" mode: one of the lemmas in this family is 
		#                             found in the lexicon
		if ( ( ( $type == $Is_morph_file and $OneLemmaInLexicon ) or
			   ( $type == $Is_sem_file and !$classes and
			     $LemmaInWordList{ $baseMemberword."#".$baseMemberpos } == $Is_word_in_term ) or
			     ( $type == $Is_sem_file and $classes and $OneLemmaInLexicon ) ) and
			     $baseMemberword and $baseMemberpos ) {

			# scans the whole family to memorize the links
			my ($baseMember, $famMembers) = split(":", $line);
			$baseMemberword =~ s/\\/\\\\/g;
			$baseMemberword =~ s/'/\\'/g;
			$baseMemberword =~ s/ +/\-/g;
			$baseMemberword =~ s/\(.*\)//g;
			my @famMembers = split( / / , $famMembers );

			# semantic families of polysemous words are not retained
			# affect also the reading of fich.morpho => families with 
			# more then 10 elements are ignored. 
			# Add 1 constraint on that type of file
			if((@famMembers <= 10 && $type == $Is_sem_file) || ($type == $Is_morph_file)) {
				# adds the whole family
				while ($famMembers =~ m/[^ ]+?\\.+?\\.*? /g) {
					my ($word, $pos, $famMember) =  split /\\/, $&;
		    
					$word =~ s/'/\\'/g;
					$word =~ s/ +/\-/g;
					$famMember =~ s/\\/\\\\/g;
					$famMember =~ s/'/\\'/g;
					$famMember =~ s/ +$//;
		    
					# Morphology: memorizes the stem of the baseMember
					# only if it is different from the current word
					if ( $type == $Is_morph_file and
					     $word."#".$pos ne $baseMemberword."#".$baseMemberpos ) {
						$stem{ $word."#".$pos } = "('".$baseMemberword."',".$baseMemberpos.")";
						}
		    
					# Semantics and "link" mode: memorizes the semantic link 
					# from the baseMember to the current word
					elsif ( $type == $Is_sem_file && !$classes ) {
						if ( !$semLink{ $baseMemberword."#".$baseMemberpos } ) {
							$semLink{ $baseMemberword."#".$baseMemberpos } = "('".$word."',".$pos.")";
							}
						else {
							$semLink{ $baseMemberword."#".$baseMemberpos } .= "|"."('".$word."',".$pos.")";
							}
						}

					# Semantics and "class" mode: memorizes the semantic link 
					# from the current word to the baseMember
					elsif ( $type == $Is_sem_file && $classes ) {
						if ( !$semLink{ $word."#".$pos } ) {
							$semLink{ $word."#".$pos } = "('".$baseMemberword."',".$baseMemberpos.")";
							}
						else {
							$semLink{ $word."#".$pos } .= "|"."('".$baseMemberword."',".$baseMemberpos.")";
							}
						}
		    
					# memorizes that the word must be written 
					if ( !($LemmaInWordList{ $word."#".$pos } == $Is_word_in_term ) ) {
						$LemmaInWordList{ $word."#".$pos } = $Is_word_in_family;
						}
					if (!($LemmaInWordList{ $baseMemberword."#".$baseMemberpos } == $Is_word_in_term ) ) {
						$LemmaInWordList{ $baseMemberword."#".$baseMemberpos } = $Is_word_in_family;
						}
					}
				# adds the whole family
				}
			# semantic families of polysemous words are not retained
			}	
		}
	close( FILEfamilies );
	}

my @newrules;
# prints the lemmas and their morph. and sem. links
while ( my ( $wordpos , $value ) = each( %LemmaInWordList ) ) {
	my ( $word , $pos ) = split( /\#/ , $wordpos );

	my $newrule;
	if( ($word ne "") && $pos ) {
		$newrule = sprintf  "Word '%s' : <cat> = %s" , $word , $pos;
		}
    
	if( $semLink{ $word."#".$pos } ) {
		$newrule .= sprintf  " <syn> = %s" , $semLink{ $word."#".$pos };
		}
	if( $stem{ $word."#".$pos } ) {
		$newrule .= sprintf  " <root> = %s" , $stem{ $word."#".$pos };
		}
	if ( $newrule !~ /^\s*\z/o ) {
		$newrule .= ".";
		push @newrules, $newrule;
		}
	}

@{$ref} = @newrules;
}


__DATA__

#
# Special French expressions that should not be split
# (from Achim Stein's "tree-tagger-fr" and 
#  Helmut Schmid's "french-abbreviatons-utf8")
#

%% FRANCAIS

Aujourd'hui
aujourd'hui
C'est-à-dire
c'est-à-dire
D'abord
d'abord
D'accord
d'accord
D'ailleurs
d'ailleurs
D'après
d'après
D'autant
d'autant
Elle-même
elle-même
Elles-mêmes
elles-mêmes
Ets.
etc.
Eux-mêmes
eux-mêmes
Garde-à-vous
garde-à-vous
Inc.
Lui-même
lui-même
M.
Mlle.
MM.
Mme.
Mr.
Mam'zelle
mam'zelle
Moi-même
moi-même
Nous-mêmes
nous-mêmes
Par-ci
par-ci
Par-là
par-là
P'tit
p'tit
Quelqu'un
quelqu'un
Rendez-vous
rendez-vous
Tél.
tél.
Toi-même
toi-même
V'là
v'là
vous-mêmes
Vous-mêmes

#
# English abbreviations (from Helmut Schmid's 
# "english-abbreviations")
#

%% ENGLISH

Adm.
Ala.
Ariz.
Ark.
Aug.
Ave.
Bancorp.
Bhd.
Brig.
Bros.
CO.
CORP.
COS.
Ca.
Calif.
Canada-U.S.
Canadian-U.S.
Capt.
Cia.
Cie.
Co.
Col.
Colo.
Conn.
Corp.
Cos.
D-Mass.
Dec.
Del.
Dept.
Dr.
Drs.
Etc.
Feb.
Fla.
Ft.
Ga.
Gen.
Gov.
Hon.
INC.
Ill.
Inc.
Ind.
Jan.
Japan-U.S.
Jr.
Kan.
Korean-U.S.
Ky.
La.
Lt.
Ltd.
Maj.
Mass.
Md.
Messrs.
Mfg.
Mich.
Minn.
Miss.
Mo.
Mr.
Mrs.
Ms.
Neb.
Nev.
No.
Nos.
Nov.
Oct.
Okla.
Ont.
Ore.
Pa.
Ph.
Prof.
Prop.
Pty.
Rep.
Reps.
Rev.
S.p.A.
Sen.
Sens.
Sept.
Sgt.
Sino-U.S.
Sr.
St.
Ste.
Tenn.
Tex.
U.S.-U.K.
U.S.-U.S.S.R.
Va.
Vt.
W.Va.
Wash.
Wis.
Wyo.
a.k.a.
a.m.
anti-U.S.
cap.
days.
etc.
ft.
i.e.
non-U.S.
office\/dept.
p.m.
president-U.S.
s.r.l.
v.
v.B.
v.w.
vs.

%% END

